"""
Gaia photometry tools.

This package provides classes and functions for transforming between the Gaia
photometric system and other photometric systems. The formulae and tables are
taken from the `Gaia science performance pages
<https://www.cosmos.esa.int/web/gaia/science-performance>`_ and from the
paper by `Jordi et al <http://adsabs.harvard.edu/abs/2010A%26A...523A..48J>`_.
"""

from . import transformations
from . import utils
