"""
Gaia simulation tools.

This package provides classes and functions for simulating the uncertainties
expected on the Gaia catalogue data. The formulae and tables are taken from the
`Gaia science performance pages
<https://www.cosmos.esa.int/web/gaia/science-performance>`_.
"""

from . import astrometric
from . import photometric
from . import spectroscopic
from . import utils
